implementation class
Annotation: Element

end;

implementation instance
Annotation
{
  <doc> The enclosed element.  </doc>
  Element element;

  <doc> Actual attributes, if any.  </doc>
  Array attributes;

  <doc> The expression (implementing this parser's semantics) following the
      {element}.  </doc>
  String expr;

  <doc> The kind of repitition of this element, or, if positive, how often.
      </doc>
  mutable int repeats;
}

<doc> Designated initializer.  </doc>
id
        init Element s
  attributes Array a
        expr String e
{
  (element, attributes, expr) = (s, a, e);

  = [super init];
}

void
  checkLL1
{
  if (element != nil)
    [element checkLL1];
}

OutputStream
  compile OutputStream s
{
  Enumerator e;
  int token;

  if (element != nil)
    {
      String condition = [isa condition [[element la] enumerator]];

      if (repeats == REPEAT_OPTIONAL)
	[s print (indent, "if ", condition)];
      else if (repeats == REPEAT_STAR)
	[s print (indent, "while ", condition)];
      else if (repeats == REPEAT_PLUS)
	[s print (indent, "do")];

      [indent add ' '];
      [s print (indent, "{")];
      [indent add ' '];

      if (expr != nil && [element outputBefore attributes])
	[s print (indent, expr)];

      if (!attributes)
	[element compile s];
      else
	[element compile s withAttributes attributes];

      if (expr != nil && ![element outputBefore attributes])
	[s print (indent, expr)];

      [indent removeElementAt 1];
      [s print (indent, "}")];
      [indent removeElementAt 1];

      if (repeats == REPEAT_PLUS)
	[s print (" while ", condition, ';')];
    }
  else if (expr != nil)
    [s print (indent, expr)];

  = s;
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (run == current_run && follow != nil && [follow equal f])
    return la;
  run = current_run;

  IntegerRangeSet s = [IntegerRangeSet new];
  follow = f;

  if (la != nil)
    [s uniteWith la];

  if (element == nil || repeats == REPEAT_OPTIONAL || repeats == REPEAT_STAR)
    [s uniteWith follow];

  if (element != nil)
    [s uniteWith [element computeLA follow]];

  boolean b = !la || ![la equal s];
  la = s;
  if (b && element != nil && (repeats == REPEAT_STAR || repeats == REPEAT_PLUS))
    [element computeLA la];

  if (!element)
    {
      first = [IntegerRangeSet new];
      [first add EPSILON];
    }
  else
    first = [element first];

  = la;
}

boolean
  isTerminal
{
  = !element ? NO : [element isTerminal];
}

OutputStream (s)
  describe-flat-rule OutputStream s
{
  if (repeats != 0)
    [self unimplemented cmd message: "non-zero repeats"];

  if (!element)
    [s print "_epsilon_"];
  else
    [element describe-flat-rule s];
}

Element
  makeFlatRules String lhs_name
pre
  !lhs_name
{
  Element ns = !element ? nil : [element makeFlatRules nil];
  if (repeats == 0)
    if (ns == element)
      = self;
    else
      = [[isa alloc] init ns attributes attributes expr expr];
  else if (repeats == REPEAT_STAR)
    {
      /* NT -> e NT; NT -> empty */
      NonTerminal nt = [NonTerminal newAnonymous];
      Sequence s1 = [[Sequence alloc] init];

      [s1 add [[isa alloc] init ns attributes attributes expr expr]];
      [s1 add nt];
      [[FlatRule alloc] init (nt, s1)];

      [[FlatRule alloc]
	init (nt, [[isa alloc] init nil attributes nil expr nil])];

      = nt;
    }
  else if (repeats == REPEAT_PLUS)
    {
      /* NT1 -> e NT2; NT2 -> e NT2; NT2 -> empty */
      NonTerminal nt1 = [NonTerminal newAnonymous];
      NonTerminal nt2 = [NonTerminal newAnonymous];

      Sequence s1 = [[Sequence alloc] init];
      Annotation ne1 = [[isa alloc] init ns attributes attributes expr expr];
      [s1 add ne1];
      [s1 add nt2];
      [[FlatRule alloc] init (nt1, s1)];

      Sequence s2 = [[Sequence alloc] init];
      // Is it OK to use NS again?
      // Tue Nov  3 20:12:23 1998, tiggr@gerbil.org
      Annotation ne2 = [[isa alloc] init ns attributes attributes expr expr];
      [s2 add ne2];
      [s2 add nt2];
      [[FlatRule alloc] init (nt2, s2)];

      [[FlatRule alloc]
	init (nt2, [[isa alloc] init nil attributes nil expr nil])];

      = nt1;
    }
  else if (repeats == REPEAT_OPTIONAL)
    {
      /* NT -> e; NT -> empty */
      NonTerminal nt = [NonTerminal newAnonymous];

      [[FlatRule alloc]
	init (nt, [[isa alloc] init ns attributes attributes expr expr])];
      [[FlatRule alloc]
	init (nt, [[isa alloc] init nil attributes nil expr nil])];

      = nt;
    }
}

OutputStream (s)
  writeFields OutputStream s
{
  = [[super writeFields s] print (" element=", element, " repeats=", repeats)];
}

<doc> Just write the {element} if this annotation effectively annotates
    nothing.  </doc>
OutputStream (s)
  write OutputStream s
{
  = (element != nil && !attributes && !expr && !repeats
     ? [element write s] : [super write s]);
}

// Yum
int
  cost
{
  = !element ? 0 : [element cost];
}

end;
